; ------------------------ MOTROT.ASM -----------------------------------
; random number generators by Agner Fog 1999
; 32-bit mode version for 80x86 and compatible microprocessors.
; Works with Window 95 and later, and possibly with other 32-bit systems.
;
; The code combines the RANROT type W and the Mother-of-All generators.
; This gives the best possible randomness for the most demanding
; applications. The resolution is 63 bits for XRandom.
;
; XRandomInit must be called before the first call to any of the random
; functions.
; XRandom returns a floating point number between 0 and 1.
; XIRandom returns an integer in the interval set by XSetInterval
;
; C++ prototypes:
; extern"C" void _cdecl XRandomInit (int seed);
; extern"C" void _cdecl XSetInterval (int min, int max);
; extern"C" double _cdecl XRandom (void);
; extern"C" int _cdecl XIRandom (void);
;
; -----------------------------------------------------------------------

; The code is optimized for the Pentium microprocessor.

.386

; define appropriate function for error-exit:
; (depends on system):
; ErrorExit equ <FatalAppExitA>

; define parameters for RANROT
JJ      equ     10            ; lag 1
KK      equ     17            ; lag 2, size of circular buffer
R1      equ     19            ; rotate count
R2      equ     27            ; rotate count
SAFE    equ      1            ; define this if you want self-test

RAND_DATA SEGMENT PARA USE32 PUBLIC 'DATA'
align   16
randp1  DT      1.5           ; used for conversion to float
        DW      0             ; alignment
p1      DD      0             ; pointer in circular buffer
p2      DD      0             ; pointer in circular buffer
randbuf DD      (2*KK) dup(?) ; circular buffer for RANROT
M0      DD      0             ; history buffer for Mother-of-All
M1      DD      0
M2      DD      0
M3      DD      0
MC      DD      0

MF3     DD      2111111111    ; factors for Mother-of-All
MF2     DD      1492
MF1     DD      1776
MF0     DD      5115

one     DD      1.0           ; constant
min     DD      0             ; used by XSetInterval
interval DD     0             ; used by XSetInterval
ALIGN   8
finterval DQ    0             ; used by XSetInterval

IFDEF   SAFE
randbufcopy DD  (4*KK) dup(?) ; double copy of randbuf initial state
RandMessage1    DB 'Random number generator not initialized',0
RandMessage2    DB 'Random number generator returned to initial state',0
ENDIF
RAND_DATA ENDS


_TEXT   SEGMENT DWORD USE32 PUBLIC 'CODE'
FLAT GROUP RAND_DATA,_TEXT
ASSUME CS:FLAT,DS:FLAT,ES:FLAT

PUBLIC XRandomInit,_XRandomInit,XSetInterval,_XSetInterval
PUBLIC XRandom,_XRandom,XIRandom,_XIRandom


IFNDEF  SAFE                    ; without self-test
XRandom PROC NEAR
_XRandom LABEL NEAR
        FLD     [randp1]        ; load random number
        FSUB    [one]           ; normalize
; RANROT-W algorithm
        MOV     EAX, [p1]       ; ring buffer pointer
        MOV     EDX, [p2]       ; ring buffer pointer
        PUSH    EBX
        NOP                     ; avoid imperfect pairing due to AGI
        MOV     EBX, [randbuf][EAX]
        MOV     ECX, [randbuf][EAX+4]

ELSE                            ; include self-test
XRandom PROC NEAR
_XRandom LABEL NEAR
        FLD     [randp1]        ; load random number
; RANROT-W algorithm:
        MOV     EAX, [p1]       ; ring buffer pointers
        MOV     ECX, [randbufcopy][0] ; first dword of initial state
        FSUB    [one]           ; normalize random number
R10:    PUSH    EBX
        MOV     EBX, [randbuf][EAX]
        CMP     ECX, EBX        ; self-test first dword
        JE      R50             ; jump if further compare needed
R20:    MOV     EDX, [p2]       ; ring buffer pointer
        MOV     ECX, [randbuf][EAX+4]
ENDIF
        ROL     EBX, R1                  ; rotate bits
        ROL     ECX, R2
        PUSH    ESI
        PUSH    EDI
        ADD     EBX, [randbuf][EDX]      ; add two dwords
        ADD     ECX, [randbuf][EDX+4]
        MOV     [randbuf][EAX], ECX      ; save in swapped order
        MOV     [randbuf][EAX+4], EBX
        SUB     EAX, 8                   ; decrement p1
        JNC     SHORT R30
        MOV     EAX, (KK-1)*8            ; wrap around p1
R30:    SUB     EDX, 8                   ; decrement p2
        JNC     SHORT R40
        MOV     EDX, (KK-1)*8            ; wrap around p2
R40:    MOV     [p1], EAX                ; save updated pointers
        MOV     [p2], EDX
; result of RANROT-W is in EBX:ECX

; Mother-of-All algorithm:
        MOV     EAX, [MF3]
        MUL     [M3]             ; X[n-4]
        MOV     ESI,EAX
        MOV     EAX, [M2]        ; X[n-3]
        MOV     EDI,EDX
        MOV     [M3],EAX
        MUL     [MF2]
        ADD     ESI,EAX
        MOV     EAX, [M1]        ; X[n-2]
        ADC     EDI,EDX
        MOV     [M2],EAX
        MUL     [MF1]
        ADD     ESI,EAX
        MOV     EAX,[M0]         ; X[n-1]
        ADC     EDI,EDX
        MOV     [M1],EAX
        MUL     [MF0]
        ADD     EAX,ESI
        ADC     EDX,EDI
        ADD     EAX,[MC]
        ADC     EDX,0
        MOV     [M0],EAX
        MOV     [MC],EDX
; combine output from RANROT and Mother-of-All
        ROL     EDX,7
        ADD     EAX,ECX
        ADD     EDX,EBX
        OR      EAX, 80000000H           ; bit 63 must be set
        MOV     DWORD PTR [randp1],EDX   ; low 32 bits of next random number
        MOV     DWORD PTR [randp1+4],EAX ; bits 32-62 of next random number
        POP     EDI
        POP     ESI
        POP     EBX
        RET

IFDEF   SAFE
R50:    ; first dword is identical. compare entire buffer
        PUSHAD
        MOV     ESI, OFFSET DS:randbuf
        MOV     EDI, OFFSET DS:randbufcopy+KK*8
        SUB     EDI, ECX                  ; buffer is rotated
        MOV     ECX, KK*2
        CLD
        REPE    CMPSD                     ; compare buffers
        JE      SHORT R60
        POPAD
        JMP     R20                       ; buffers are different. continue
R60:    ; error found. initial state is repeated
IFDEF   ErrorExit
        PUSH    OFFSET DS:RandMessage1
        MOV     EAX, [p1]
        CMP     EAX, [p2]
        JE      SHORT R70
        PUSH    OFFSET DS:RandMessage2
R70:    PUSH    0
        CALL    ErrorExit   ; name of abort procedure depends on system
        EXTRN   ErrorExit:NEAR
ELSE
        ; ErrorExit function not defined. Make division by 0 error
        DIV     ECX
ENDIF
ENDIF
XRandom ENDP

XSetInterval PROC NEAR
_XSetInterval LABEL NEAR
        MOV     EAX, [ESP+4]            ; min
        MOV     EDX, [ESP+8]            ; max
        CMP     EAX, EDX
        JL      RSI1
        XCHG    EAX, EDX                ; make sure min <= max
RSI1:   MOV     [min], EAX
        SUB     EDX, EAX
        INC     EDX
        MOV     [interval], EDX         ; max-min
        FILD    [interval]
        FSTP    [finterval]
        RET
XSetInterval ENDP

XIRandom PROC   NEAR
_XIRandom LABEL NEAR
        CALL    XRandom
        FMUL    [finterval]
        SUB     ESP, 4
        FISTP   DWORD PTR [ESP]
      ; WAIT                    ;(only needed for old processors)
        POP     EAX
        CMP     EAX,[interval]
        ADC     EAX,-1          ; make sure value < interval
        ADD     EAX,[min]
        RET
XIRandom ENDP

XRandomInit PROC NEAR
_XRandomInit LABEL NEAR
        MOV     EAX, [ESP+4]    ; seed
        CMP     EAX, 1
        SBB     EAX, 0          ; make sure seed != 0
        XOR     ECX, ECX
        ; make random numbers and put them into buffers
R80:    MOV     EDX, EAX
        SHL     EAX, 13
        XOR     EDX, EAX
        MOV     EAX, EDX
        SHR     EDX, 17
        XOR     EAX, EDX
        MOV     EDX, EAX
        SHL     EDX, 5
        XOR     EAX, EDX
        MOV     [randbuf][ECX*4], EAX
        INC     ECX
        CMP     ECX, KK*2 + 5
        JB      R80
        FLD1
        FSTP    [randp1]        ; initialize bit 64-79
        MOV     [p1], 0         ; initialize buffer pointers
        MOV     [p2], JJ*8
IFDEF   SAFE
        PUSHAD                  ; make 2 copies of initial state
        MOV     ECX, KK*2
        MOV     ESI, OFFSET DS:randbuf
        MOV     EDI, OFFSET DS:randbufcopy
        PUSH    ECX
        PUSH    ESI
        CLD
        REP     MOVSD
        POP     ESI
        POP     ECX
        REP     MOVSD
        INC     [randbufcopy][0]   ; avoid error message
        CALL    XRandom            ; prepare first random number
        FSTP    ST
        DEC     [randbufcopy][0]
        POPAD
ELSE
        CALL    XRandom            ; prepare first random number
        FSTP    ST
ENDIF
        PUSH    EDI
        MOV     EDI, 96
R90:    CALL    XRandom
        FSTP    ST
        DEC     EDI
        JNZ     R90
        POP     EDI                                                
        RET     0                  ; RET 4 if not _cdecl calling
XRandomInit ENDP

_TEXT   ENDS
        END

; To optimize for Pentium II processor: Remove the NOP and replace
; the conditional jumps to R30 and R40 with conditional moves.


; Example of calling from C++ program:

; #include "arandom.h"
; int i;
; XRandomInit (time(0)); XSetInterval(1,10);
; for (i=0; i<100; i++) {
;   printf ("\n%14.10f  %2i", XRandom(), XIRandom);}
