/************************* RANROTB.C ******************** AgF 1999-03-03 *
*  Random Number generator 'RANROT' type B                               *
*                                                                        *
*  This is a lagged-Fibonacci type of random number generator with       *
*  rotation of bits.  The algorithm is:                                  *
*  X[n] = ((X[n-j] rotl r1) + (X[n-k] rotl r2)) modulo 2^b               *
*                                                                        *
*  The last k values of X are stored in a circular buffer named          *
*  randbuffer.                                                           *
*                                                                        *
*  This version works with any integer size: 16, 32, 64 bits etc.        *
*  The integers must be unsigned. The resolution depends on the integer  *
*  size.                                                                 *
*                                                                        *
*  Note that the function RanrotAInit must be called before the first    *
*  call to RanrotA or iRanrotA                                           *
*                                                                        *
*  The theory of the RANROT type of generators is described at           *
*  www.agner.org/random/ranrot.htm                                       *
*                                                                        *
*************************************************************************/

/* names of header files may differ */
#include <stdio.h>
#include <conio.h>
#include <mem.h>
#include <math.h>
#include <time.h>


/* define desired integer type */
typedef  unsigned int my_uint;

/* If your system doesn't have a rotate function for the chosen integer type
   then define it thus:                                                      */
my_uint rotl (my_uint x, my_uint r) {
  return (x << r) | (x >> (sizeof(x)*8-r));}

/* define parameters (R1 and R2 must be smaller than the integer size): */
#define KK  17
#define JJ  10
#define R1   5
#define R2   3

/* global variables */
my_uint randbuffer[KK];  /* history buffer */
int r_p1, r_p2;          /* indexes into history buffer */
float scale;             /* 2^(- integer size) */


/* returns a random number between 0 and 1 */
double RanrotA() {
  my_uint x;
  /* generate next random number */
  x = randbuffer[r_p1] = rotl(randbuffer[r_p2], R1) + rotl(randbuffer[r_p1], R2);
  /* rotate list pointers */
  if (--r_p1 < 0) r_p1 = KK - 1;
  if (--r_p2 < 0) r_p2 = KK - 1;
  /* conversion to float */
  return x * scale;}


/* get integer random number in interval from min to max */
int iRanrotA(int min, int max) {
  int i, r;
  i = max - min + 1;
  r = i * RanrotA();
  if (r >= i) r = i;
  return min + r;}


/* this function initializes the random number generator.      */
/* Must be called before the first call to RanrotA or iRanrotA */
void RanrotAInit (my_uint seed) {
  int i;

  /* put semi-random numbers into the buffer */
  for (i=0; i<KK; i++) {
    randbuffer[i] = seed;
    seed = rotl(seed,5) + 97;}

  /* initialize pointers to circular buffer */
  r_p1 = 0;  r_p2 = JJ;

  /* randomize */
  for (i = 0;  i < 300;  i++) RanrotA();

  /* compute 2^(- integer size) */
  scale = ldexp(1, -8*sizeof(my_uint));
}


/**************************************************************************/
/* Example showing how to use the random number generator: */
void main() {
  int i;

  /* initialize, using time as random seed */
  RanrotAInit(time(0));

  /* generate 10 random floating point numbers and 10 random integers */
  for (i=0; i<20; i++) {
    printf("\n%14.10f   %2.0f", RanrotA(), (double)iRanrotA(1,10));}

  getch();  /* wait for user to press any key */
  }
